
function [interped,denoised] = deNoiseAndInterp( imgH, imgL, smoothSigma, varargin )
  % [ interped, denoised ] = deNoiseAndInterp( imgH, imgL, smoothSigma [, 
  %   'lambda', lambda, 'N', N, 'ws', ws, 'verbose', true/false ] )
  %
  % Inputs:
  % imgH - the high resolution image
  % imgL - the low resolution image
  % smoothSigma - the standard deviation of the Gaussian blur kernel
  %
  % Optional Inputs:
  % lambda - regularization parameter
  % N - the number of iterations that Chambolle-Pock will employ
  % ws - weights of the weighted norm
  % verbose - if set, shows informative information
  %
  % Written by Nicholas Dwork, Copyright 2020

  p = inputParser;
  p.addRequired( 'imgH', @isnumeric );
  p.addRequired( 'imgL', @isnumeric );
  p.addRequired( 'sigma', @ispositive );
  p.addParameter( 'lambda', 1d3, @isnonnegative );  % joint sparsity regularization parameter
  p.addParameter( 'N', 300, @isnumeric );
  p.addParameter( 'ws', 1, @isnumeric );
  p.addParameter( 'verbose', 1, @(x) isnumeric(x) || islogical(x) );
  p.parse( imgH, imgL, smoothSigma, varargin{:} );
  lambda = p.Results.lambda;
  N = p.Results.N;
  ws = p.Results.ws;
  verbose = p.Results.verbose;

  maxImgH = max( imgH(:) );  imgH = imgH / maxImgH;
  maxImgL = max( imgL(:) );  imgL = imgL / maxImgL;

  nH = numel( imgH );  sH = size( imgH );
  nL = numel( imgL );  sL = size( imgL );

  interpedL = imresize( imgL, sH, 'bicubic' );
  %ws = imresize( imgL, sH, 'bilinear' );

  lambda = lambda / nH;

  k = ceil( max( 5 * smoothSigma, 3 ) );  % smoothing kernel size
  k( mod( k, 2 ) == 0 ) = k( mod( k, 2 ) == 0 ) + 1;

  function out = applyDB( x, op )
    if nargin < 2 || strcmp( op, 'notransp' )
      % Here, x is the interpolated image IL
      smoothX = smoothImg( x, k, 'gaussian', smoothSigma );
      out = downsize2( smoothX, sL, 'notransp' );
    elseif strcmp( op, 'transp' )
      DTx = downsize2( x, sH, 'transp' );
      out = smoothImg( DTx, k, 'gaussian', smoothSigma, 'op', 'transp' );
        % Note: could be made faster by only accounting for non-zero values
    else
      error( 'Unrecognized op value' );
    end
  end


  function out = K( x, op )
    if nargin < 2 || strcmp( op, 'notransp' )
      IH = reshape( x( 1 : nH ), sH );
      IL = reshape( x( nH + 1 : end ), sH );

      DBIL = applyDB( IL );
      GIH = computeGradient( IH );
      GIL = computeGradient( IL );

      wGIH = bsxfun( @times, GIH, ws );
      wGIL = bsxfun( @times, GIL, ws );

      out = [ IH(:) / sqrt( nH );  ...
              DBIL(:) / sqrt( nL ); ...
              wGIH(:); ...
              wGIL(:); ];

    elseif strcmp( op, 'transp' )
      x1 = reshape( x( 1 : nH ), sH ) / sqrt( nH );
      x2 = reshape( x( nH + 1 : nH + nL ), sL ) / sqrt( nL );
      x3 = reshape( x( nH + nL + 1 : 3 * nH + nL ), [ sH 2 ] );
      x4 = reshape( x( 3 * nH + nL + 1 : end ), [ sH 2 ] );

      wx3 = bsxfun( @times, x3, ws );
      wx4 = bsxfun( @times, x4, ws );

      BTDTx2 = applyDB( x2, 'transp' );
      GTx3 = computeGradient( wx3, 'transp' );
      GTx4 = computeGradient( wx4, 'transp' );

      out = [ x1(:);  BTDTx2(:); ] + [ GTx3(:);  GTx4(:); ];
    end
  end

  checkAdjoints = 0;
  if checkAdjoints == 1
    if ~checkAdjoint( imgH, @applyDB, 'y', imgL ), error( 'applyDB adjoint test failed' ); end
    if ~checkAdjoint( rand( 2*nH, 1 ), @K ), error( 'K adjoint test failed' ); end
  end


  f = @(x) 0;

  b1 = imgH(:) / sqrt( nH );
  b2 = imgL(:) / sqrt( nL );
  b = [ b1; b2; ];

  function out = g( x )
    x1 = x( 1 : nH );
    x2 = x( nH + 1 : nH  + nL );
    x3 = x( nH + nL + 1 : 3 * nH + nL );  x3 = reshape( x3, [ sH 2 ] );
    x4 = x( 3 * nH + nL + 1 : end );      x4 = reshape( x4, [ sH 2 ] );

    out = 0.5 * norm( x1(:) - b1, 2 )^2 + ...
          0.5 * norm( x2(:) - b2, 2 )^2 + ...
          lambda * normL2L1( cat( 3, x3, x4 ) );

    %disp( [ '   ', num2str( 0.5 * norm( x1(:) - b1, 2 )^2 ), ...
    %    ',  ', num2str( 0.5 * norm( x2(:) - b2, 2 )^2 ), ...
    %    ',  ', num2str( lambda * normL2L1( cat( 3, x3, x4 ) ) ), ...
    %    ',  total: ', num2str(out) ] );
  end

  proxtf = @(x,~) min( max( x, 0), 1 );
  %proxtf = @(x,~) x;

  proxg1Conj = @(x,sigma) ( x - sigma * b ) / ( 1 + sigma );


  function out = proxg2Conj( x, ~ )
    if lambda == 0
      out = zeros( size(x) );
      return;
    end
    
    x = reshape( x, [ sH 4 ] );
    nx = norms( x, 2 );

    scalings = ones( size( nx ) );
    scalings( nx > lambda ) = lambda ./ scalings( nx > lambda );

    out = bsxfun( @times, x, scalings );
  end


  function out = proxgConj( x, sigma )
    n1 = nH + nL;
    x1 = x( 1 : n1 );
    out1 = proxg1Conj( x1, sigma );

    x2 = x( n1 + 1 : end );
    out2 = proxg2Conj( x2, sigma );

    out = [ out1(:); out2(:); ];
  end


  x0 = [ imgH(:); interpedL(:); ];

  alg = 'pdhgWLS';
  if strcmp( 'pdhg', alg )
    normK = powerIteration( @K, x0(:), 'tolerance', 0 );
    tau = 1d-3 / normK;
    [xFinal,objValues] = chambollePock( x0(:), proxtf, @proxgConj, tau, ...
      'A', @K, 'f', f, 'g', @g, 'N', N, 'normA', normK, ...
      'printEvery', 50, 'verbose', verbose );

  elseif strcmp( 'pdhgWLS', alg )
    tau = 1;
    beta = 2;
    [xFinal,objValues] = pdhgWLS( x0(:), proxtf, @proxgConj, 'A', @K, ...
      'N', N, 'tau', tau, 'f', f, 'g', @g, 'beta', beta, ...
      'printEvery', 10, 'verbose', verbose );
  end

  denoised = reshape( xFinal( 1 : nH ), sH );
  interped = reshape( xFinal( nH+1 : end ), sH );


  if verbose == true
    imshowscale( [ imresize(imgL,sH), interpedL, interped, imgH, denoised ], 5, 'range', imgH );
    figure; imshowscale( imresize(imgL,sH,'nearest'), 5 );  titlenice( 'imgL' );
    figure; imshowscale( interpedL, 5 );  titlenice( 'linearly interped' );
    figure; imshowscale( interped, 5 );  titlenice( 'interped' );
    figure; imshowscale( imgH, 5 );  titlenice( 'imgH' );
    figure; imshowscale( denoised, 5 );  titlenice( 'denoised' );
    figure; plotnice( objValues );
  end

  interped = interped * maxImgL;
  denoised = denoised * maxImgH;
end

